from .beta_loss import BetaCrossEntropyLoss
from .binomial_loss import BinomialCrossEntropyLoss
from .custom_targets_loss import CustomTargetsCrossEntropyLoss
from .exponential_loss import ExponentialRegularisedCrossEntropyLoss
from .general_triangular_loss import GeneralTriangularCrossEntropyLoss
from .mceloss import MCELoss
from .mcewkloss import MCEAndWKLoss
from .ordinal_ecoc_distance_loss import OrdinalECOCDistanceLoss
from .poisson_loss import PoissonCrossEntropyLoss
from .triangular_loss import TriangularCrossEntropyLoss
from .wkloss import WKLoss

__all__ = [
    "CustomTargetsCrossEntropyLoss",
    "BetaCrossEntropyLoss",
    "PoissonCrossEntropyLoss",
    "BinomialCrossEntropyLoss",
    "ExponentialRegularisedCrossEntropyLoss",
    "WKLoss",
    "MCELoss",
    "MCEAndWKLoss",
    "OrdinalECOCDistanceLoss",
    "TriangularCrossEntropyLoss",
    "GeneralTriangularCrossEntropyLoss",
]



from .clm import CLM
from .ordinal_fully_connected import (
    ResNetOrdinalFullyConnected,
    VGGOrdinalFullyConnected,
)
from .stick_breaking_layer import StickBreakingLayer

__all__ = [
    "CLM",
    "ResNetOrdinalFullyConnected",
    "VGGOrdinalFullyConnected",
    "StickBreakingLayer",
]



from .metrics import (
    accuracy_off1,
    amae,
    gmsec,
    minimum_sensitivity,
    mmae,
    write_array_to_file,
    write_metrics_dict_to_file,
)

__all__ = [
    "gmsec",
    "minimum_sensitivity",
    "accuracy_off1",
    "amae",
    "mmae",
    "write_array_to_file",
    "write_metrics_dict_to_file",
]



from typing import Optional

import torch
from torch import Tensor

from .mceloss import MCELoss
from .wkloss import WKLoss


class MCEAndWKLoss(torch.nn.modules.loss._WeightedLoss):
    """
    The loss function integrates both MCELoss and WKLoss, concurrently minimising
    error distances while preventing the omission of classes from predictions.

    Parameters
    ----------
    num_classes : int
        Number of classes
    C: float, defaul=0.5
        Weights the WK loss (C) and the MCE loss (1-C). Must be between 0 and 1.
    wk_penalization_type : str, default='quadratic'
        The penalization type of WK loss to use (quadratic or linear).
        See WKLoss for more details.
    weight : Optional[Tensor], default=None
        A manual rescaling weight given to each class. If given, has to be a Tensor
        of size `J`, where `J` is the number of classes.
        Otherwise, it is treated as if having all ones.
    reduction : str, default='mean'
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` |
        ``'sum'``. ``'none'``: no reduction will be applied, ``'mean'``: the sum of
        the output will be divided by the number of elements in the output,
        ``'sum'``: the output will be summed.
    """

    def __init__(
        self,
        num_classes: int,
        C: float = 0.5,
        wk_penalization_type: str = "quadratic",
        weight: Optional[Tensor] = None,
        reduction: str = "mean",
    ) -> None:
        super().__init__(
            weight=weight, size_average=None, reduce=None, reduction=reduction
        )

        self.num_classes = num_classes
        self.C = C
        self.wk_penalization_type = wk_penalization_type

        if weight is not None and weight.shape != (num_classes,):
            raise ValueError(
                f"Weight shape {weight.shape} is not compatible"
                + "with num_classes {num_classes}"
            )

        if C < 0 or C > 1:
            raise ValueError(f"C must be between 0 and 1, but is {C}")

        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(
                f"Reduction {reduction} is not supported."
                + " Please use 'mean', 'sum' or 'none'"
            )

        self.wk = WKLoss(
            self.num_classes,
            penalization_type=self.wk_penalization_type,
            weight=weight,
        )
        self.mce = MCELoss(self.num_classes, weight=weight)

    def forward(self, y_true: torch.Tensor, y_pred: torch.Tensor):
        """
        Parameters
        ----------
        y_true : torch.Tensor
            Ground truth labels
        y_pred : torch.Tensor
            Predicted labels

        Returns
        -------
        loss : torch.Tensor
            The weighted sum of MCE and QWK loss
        """

        wk_result = self.wk(y_true, y_pred)
        mce_result = self.mce(y_true, y_pred)

        return self.C * wk_result + (1 - self.C) * mce_result



import numpy as np
from scipy.stats import binom


def get_binomial_soft_labels(J):
    """Get soft labels for the binomial distribution for ``J`` classes or splits
    using the approach described in :footcite:t:`liu2020unimodal`.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the binomial
    probability function for the interval boundaries. The probability for the first
    interval is computed as the value of the binomial probability function for the first
    interval boundary.

    The binomial distributions employed are denoted as :math:`\\text{b}(k, n-1, p)` where
    :math:`k` is given by the order of the class for which the probability is computed,
    and :math:`p` is given by :math:`0.1 + (0.9-0.1) / (n-1) * j` where :math:`j` is
    is the order of the target class.

    Parameters
    ----------
    J : int
            Number of classes or splits.

    Raises
    ------
    ValueError
            If ``J`` is not a positive integer greater than 1.

    Returns
    -------
    probs : 2d array-like of shape (J, J)
            Matrix of probabilities where each row represents the true class
            and each column the probability for class ``j``.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_binomial_soft_labels
    >>> get_binomial_soft_labels(5)
    array([[6.561e-01, 2.916e-01, 4.860e-02, 3.600e-03, 1.000e-04],
            [2.401e-01, 4.116e-01, 2.646e-01, 7.560e-02, 8.100e-03],
            [6.250e-02, 2.500e-01, 3.750e-01, 2.500e-01, 6.250e-02],
            [8.100e-03, 7.560e-02, 2.646e-01, 4.116e-01, 2.401e-01],
            [1.000e-04, 3.600e-03, 4.860e-02, 2.916e-01, 6.561e-01]])
    """

    if J < 2 or not isinstance(J, int):
        raise ValueError(f"{J=} must be a positive integer greater than 1")

    params = {}

    params["4"] = np.linspace(0.1, 0.9, 4)
    params["5"] = np.linspace(0.1, 0.9, 5)
    params["6"] = np.linspace(0.1, 0.9, 6)
    params["7"] = np.linspace(0.1, 0.9, 7)
    params["8"] = np.linspace(0.1, 0.9, 8)
    params["10"] = np.linspace(0.1, 0.9, 10)
    params["12"] = np.linspace(0.1, 0.9, 12)
    params["14"] = np.linspace(0.1, 0.9, 14)

    probs = []

    for true_class in range(0, J):
        probs.append(binom.pmf(np.arange(0, J), J - 1, params[str(J)][true_class]))

    return np.array(probs)



import torch
import torch.nn as nn


class HybridDropoutContainer(nn.Module):
    """Container for the ``HybridDropout`` module. This container is used to set the
    targets of the batch in the HybridDropout module.

    Parameters
    ----------
    model : torch.nn.Module
        Model to be wrapped.
    """

    def __init__(self, model):
        super(HybridDropoutContainer, self).__init__()
        self.model = model

    def forward(self, x):
        """Forward pass of the ``HybridDropoutContainer`` module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Output tensor
        """

        return self.model(x)

    def set_targets(self, targets):
        """
        Set the targets of the batch in the ``HybridDropout`` module.

        Parameters
        ----------
        targets : torch.Tensor
            Targets of the batch

        Example
        -------
        >>> from dlordinal.dropout import HybridDropoutContainer
        >>> from torchvision.models import resnet18
        >>> model = resnet18(weights='IMAGENET1K_V1')
        >>> model.fc = nn.Sequential(
                nn.Linear(model.fc.in_features, 256),
                HybridDropout(),
                nn.Linear(256, num_classes),
            )
        >>> batch_targets = torch.tensor([...])
        >>> model = HybridDropoutContainer(model)
        >>> model.set_targets(targets)
        """

        for module in self.model.modules():
            if isinstance(module, HybridDropout):
                module.batch_targets = targets


class HybridDropout(nn.Module):
    """Implements a hybrid dropout methodology by :footcite:t:`berchez2024fusion` which
    mix a standard dropout with an ordinal dropout. The ordinal dropout is based on the
    correlation between the activation values of the neuron and the target labels
    of the dataset.

    To use this module, you must wrap your model with the ``HybridDropoutContainer``
    module

    Parameters
    ----------
    p : float
        Probability of an element to be zeroed. Default: 0.5
    beta : float
        Weight of the ordinal dropout. Default: 0.1
    batch_targets : torch.Tensor
        Targets of the batch. Default: None

    Raises
    ------
    ValueError
        If ``p`` is not a probability.
    """

    def __init__(self, p: float = 0.5, beta: float = 0.1):
        super(HybridDropout, self).__init__()
        self.p = p
        self.beta = beta
        if self.p < 0 or self.p > 1:
            raise ValueError("p must be a probability")

    def forward(self, x):
        """Forward pass of the HybridDropout module just during training. The module
        calculates the correlation between the activation values of the neuron and the
        target labels of the dataset. Then, it calculates the ordinal probabilities
        and the mask for the dropout.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor

        Raises
        ------
        ValueError
            If there are NaN values in the tensor.
        ValueError
            If the batch targets have not been set.

        Returns
        -------
        torch.Tensor
            Output tensor
        """
        if self.training:
            if torch.isnan(x).any():
                raise ValueError("Nan values in the tensor")

            if hasattr(self, "batch_targets"):
                targets = self.batch_targets
            else:
                raise ValueError(
                    "Batch targets have not been set. Use"
                    " HybridDropoutContainer.set_targets() to set the targets."
                )

            targets = torch.reshape(targets, (1, targets.shape[0]))

            correlation_list = []

            # Pearson's correlation calculation for each neuron
            for neuron in range(0, x.shape[1]):
                patterns = x[:, neuron]
                patterns = torch.reshape(patterns, (1, x.shape[0]))
                concat = torch.cat((patterns, targets), 0)
                corr = torch.corrcoef(concat)
                correlation_list.append(float(corr[0, 1]))

            correlations = torch.Tensor(correlation_list)

            # Scale of the correlation matrix
            correlations = 1 + correlations
            correlations = correlations / 2
            correlations = torch.nan_to_num(correlations)

            # correlations = correlations.to(x.device)

            # Get ordinal probabilities
            ordinal_prob = 1 - correlations

            # Mask creation: the first summand is the one related to ordinal dropout and
            # the second summand is the standard dropout.
            probabilities = (self.beta * ordinal_prob) + ((1 - self.beta) * self.p)
            mask = torch.empty(x.size()[1]).uniform_(0, 1) >= probabilities
            mask = mask.to(x.device)

            # Normalisation
            no_zeros = int(torch.count_nonzero(mask))
            total_neurons = mask.shape[0]
            zeros = total_neurons - no_zeros
            probability = zeros / total_neurons

            mask = torch.reshape(mask, (1, mask.shape[0]))
            mask = mask.repeat(x.shape[0], 1)

            return x.mul(mask) * probability
        else:
            return x



import numpy as np
from scipy.special import softmax
from scipy.stats import poisson


def get_poisson_soft_labels(J):
    """Get soft labels using poisson distributions for ``J`` classes or splits using the
    methodology described in :footcite:t:`liu2020unimodal`.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the poisson
    probability function for the interval boundaries. The probability for the first
    interval is computed as the value of the poisson probability function for the first
    interval boundary. Then, a softmax function is applied to each row of the resulting
    matrix to obtain valid probabilities.

    The poisson probability function is denoted as :math:`\\text{p}(k, \\lambda)`
    where :math:`k` is given by the order of the class for which the probability is
    computed, and :math:`\\lambda` is given by :math:`k` where :math:`k` is the order
    of the target class.

    Parameters
    ----------
    J : int
        Number of classes or splits.

    Raises
    ------
    ValueError
        If ``J`` is not a positive integer greater than 1.

    Returns
    -------
    probs : 2d array-like of shape (J, J)
        Matrix of probabilities where each row represents the true class
        and each column the probability for class j.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_poisson_soft_labels
    >>> get_poisson_soft_labels(5)
    array([[0.23414552, 0.23414552, 0.19480578, 0.17232403, 0.16457916],
        [0.18896888, 0.21635436, 0.21635436, 0.19768881, 0.18063359],
        [0.17822335, 0.19688341, 0.21214973, 0.21214973, 0.20059378],
        [0.17919028, 0.18931175, 0.20370191, 0.21389803, 0.21389803],
        [0.18400408, 0.18903075, 0.19882883, 0.21031236, 0.21782399]])

    """

    if J < 2 or not isinstance(J, int):
        raise ValueError(f"{J=} must be a positive integer greater than 1")

    probs = []

    for true_class in range(1, J + 1):
        probs.append(poisson.pmf(np.arange(0, J), true_class))

    return softmax(np.array(probs), axis=1)



import numpy as np


def get_intervals(n):
    """Get n evenly-spaced intervals in :math:`[0,1]`.

    Parameters
    ----------
    n : int
        Number of intervals.

    Returns
    -------
    intervals: list
        List of intervals.
    """

    points = np.linspace(1e-9, 1 - 1e-9, n + 1)
    intervals = []
    for i in range(0, points.size - 1):
        intervals.append((points[i], points[i + 1]))

    return intervals


def triangular_cdf(x: float, a: float, b: float, c: float):
    """
    Triangular distribution CDF.

    Parameters
    ----------
    x : float
        Value.
    a : float
        Lower bound.
    b : float
        Upper bound.
    c : float
        Mode.
    """
    if x <= a:
        return 0
    elif a < x < c:
        return pow(x - a, 2) / ((b - a) * (c - a))
    elif c < x < b:
        return 1 - pow(b - x, 2) / ((b - a) * (b - c))
    else:  # b <= x
        return 1



from typing import Optional

import numpy as np
import torch
from torch import Tensor


class OrdinalECOCDistanceLoss(torch.nn.Module):
    """Ordinal ECOC distance loss from :footcite:t:`barbero2023error` for use
    with :class:`dlordinal.models.OBDECOCModel`. Computes the MSE loss
    between the output of the model (class threshold probabilities) and
    the ideal output vector for each class.

    Parameters
    ----------
    num_classes : int
        Number of classes
    weights : Optional[torch.Tensor]
        Optional weighting for each class
    """

    target_class: Tensor
    weights: Optional[Tensor]

    def __init__(self, num_classes: int, weights: Optional[Tensor] = None) -> None:
        super().__init__()
        target_class = np.ones((num_classes, num_classes - 1), dtype=np.float32)
        target_class[np.triu_indices(num_classes, 0, num_classes - 1)] = 0.0
        target_class = torch.tensor(target_class, dtype=torch.float32)
        self.register_buffer("target_class", target_class)
        self.mse = torch.nn.MSELoss(reduction="sum" if weights is None else "none")
        if weights is not None:
            self.register_buffer("weights", weights)
        else:
            self.weights = None

    def forward(self, input, target):
        """
        Parameters
        ----------
        input : torch.Tensor
            Predicted probabilities for each class threshold: :math:`P(y > q)`
        target : torch.Tensor
            Ground truth labels

        Returns
        -------
        loss : torch.Tensor
        """
        target_vector = self.target_class[target]

        if self.weights is None:
            return self.mse(input, target_vector)
        else:
            weights = self.weights[target]
            return (self.mse(input, target_vector).sum(dim=1) * weights).sum()



from typing import Optional

import torch
from torch import Tensor


class MCELoss(torch.nn.modules.loss._WeightedLoss):
    """
    Per class mean squared error loss function. Computes the mean squared error for each
    class and reduces it using the specified `reduction`.

    Parameters
    ----------
    num_classes : int
        Number of classes.

    weight : Optional[Tensor], default=None
        A manual rescaling weight given to each class. If given, has to be a Tensor
        of size `J`, where `J` is the number of classes.
        Otherwise, it is treated as if having all ones.

    reduction : str, default='mean'
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` |
        ``'sum'``. ``'none'``: no reduction will be applied, ``'mean'``: the sum of
        the output will be divided by the number of elements in the output,
        ``'sum'``: the output will be summed.
    """

    def __init__(
        self,
        num_classes: int,
        weight: Optional[Tensor] = None,
        reduction: str = "mean",
    ) -> None:
        super().__init__(
            weight=weight, size_average=None, reduce=None, reduction=reduction
        )

        self.num_classes = num_classes

        if weight is not None and weight.shape != (num_classes,):
            raise ValueError(
                f"Weight shape {weight.shape} is not compatible"
                + "with num_classes {num_classes}"
            )

        if reduction not in ["mean", "sum", "none"]:
            raise ValueError(
                f"Reduction {reduction} is not supported."
                + " Please use 'mean', 'sum' or 'none'"
            )

    def compute_per_class_mse(self, input: torch.Tensor, target: torch.Tensor):
        """
        Computes the MSE for each class independently.

        Parameters
        ----------
        input : torch.Tensor
            Predicted labels
        target : torch.Tensor
            Ground truth labels

        Returns:
        --------
        mses : torch.Tensor
            MSE values
        """

        if input.shape != target.shape:
            target = torch.nn.functional.one_hot(target, num_classes=self.num_classes)

        if input.shape != target.shape:
            raise ValueError(
                f"Input shape {input.shape} is not compatible with target shape "
                + f"{target.shape}"
            )

        # Compute the squared error for each class
        per_class_se = torch.pow(target - input, 2)

        # Apply class weights if defined
        if self.weight is not None:
            tiled_weight = torch.tile(self.weight, (per_class_se.shape[0], 1))
            per_class_se = per_class_se * tiled_weight

        # Compute the mean squared error for each class
        per_class_mse = torch.mean(per_class_se, dim=0)

        return per_class_mse

    def forward(self, input: torch.Tensor, target: torch.Tensor):
        """
        Parameters
        ----------
        input : torch.Tensor
            Predicted labels
        target : torch.Tensor
            Ground truth labels

        Returns
        -------
        reduced_mse : torch.Tensor
            MSE per class reduced using the specified `reduction`. If reduction is `none`,
            then a tensor with the MSE for each class is returned.
        """

        input = torch.nn.functional.softmax(input, dim=1)
        target_oh = torch.nn.functional.one_hot(target, num_classes=self.num_classes)

        per_class_mse = self.compute_per_class_mse(input, target_oh)

        if self.reduction == "mean":
            reduced_mse = torch.mean(per_class_mse)
        elif self.reduction == "sum":
            reduced_mse = torch.sum(per_class_mse)
        else:
            reduced_mse = per_class_mse

        return reduced_mse



from typing import Optional, Tuple

import numpy as np
import pandas as pd
import torch
from numpy.typing import ArrayLike
from sklearn.preprocessing import StandardScaler
from torch.utils.data import Dataset


class FeatureDataset(Dataset):
    """Dataset torch implementation for a standard dataset that contains several features
    that are organised in a tabular way in a csv file. The last column is the target
    variable.

    Example
    -------
    >>> train_data = FeatureDataset("train.csv")
    >>> train_data.normalize_X()
    >>> train_data.normalize_y()
    >>> train_loader = DataLoader(train_data, batch_size=32, shuffle=True)
    >>> for X, y in train_loader:
    >>>     print(X.shape, y.shape)
    >>> test_data = FeatureDataset("test.csv")
    >>> test_data.normalize_X(train_data.X_mean, train_data.X_scale)
    >>> test_data.normalize_y(train_data.y_mean, train_data.y_scale)
    >>> test_loader = DataLoader(test_data, batch_size=32, shuffle=False)
    >>> for X, y in test_loader:
    >>>     print(X.shape, y.shape)
    """

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename : str
            Path to the csv file containing the dataset.
        """

        df = pd.read_csv(filename)
        X = df.iloc[:, :-1].values
        y = df.iloc[:, -1].values

        self.X_mean: float = 0.0
        self.X_scale: float = 0.0
        self.y_mean: float = 0.0
        self.y_scale: float = 0.0

        self.X = torch.tensor(X, dtype=torch.float)
        self.targets = torch.tensor(y, dtype=torch.float)
        self.classes = torch.tensor(np.unique(y), dtype=torch.float)

    def get_valid_shape_array(self, v: ArrayLike):
        """Convert the input ArrayLike object to a 2D numpy array with shape (n, 1)
        if it is a 1D array.

        Parameters
        ----------
        v : ArrayLike
            Input array.

        Returns
        -------
        v : np.ndarray
            2D numpy array with shape (n, 1).
        """

        if isinstance(v, pd.Series):
            v = v.values  # type: ignore
        if len(v.shape) == 1:  # type: ignore
            v = np.reshape(v, (-1, 1))
        return v

    def _get_normalized(
        self,
        v: np.ndarray,
        mean: Optional[ArrayLike] = None,
        scale: Optional[ArrayLike] = None,
    ) -> Tuple[np.ndarray, ArrayLike, ArrayLike]:
        sc = StandardScaler()
        if mean is not None and scale is not None:
            sc.mean_ = mean
            sc.scale_ = scale
        else:
            sc.fit(v)
        return sc.transform(v), sc.mean_, sc.scale_

    def normalize_X(
        self, mean: Optional[ArrayLike] = None, scale: Optional[ArrayLike] = None
    ):
        """Standardize the features of the dataset.
        If mean and scale are not provided, they are computed from the dataset.
        If they are provided, they are used to standardize the dataset.

        Parameters
        ----------
        mean : array-like, default=None
            Mean of the dataset.
        scale : array-like, default=None
            Scale of the dataset.

        Returns
        -------
        self: FeatureDataset
            The dataset with standardized features.

        Example
        -------
        >>> train_data = FeatureDataset("train.csv")
        >>> train_data.normalize_X()
        >>> test_data = FeatureDataset("test.csv")
        >>> test_data.normalize_X(train_data.X_mean, train_data.X_scale)
        """

        self.X, self.X_mean, self.X_scale = self._get_normalized(
            self.get_valid_shape_array(self.X), mean, scale
        )
        self.X = torch.tensor(self.X, dtype=torch.float)
        return self

    def normalize_y(self, mean: ArrayLike = None, scale: ArrayLike = None):
        """Standardize the target variable of the dataset.
        If mean and scale are not provided, they are computed from the dataset.
        If they are provided, they are used to standardize the dataset.

        Parameters
        ----------
        mean : array-like, default=None
            Mean of the dataset.
        scale : array-like, default=None
            Scale of the dataset.

        Returns
        -------
        self: FeatureDataset
            The dataset with standardized target variable.

        Example
        -------
        >>> train_data = FeatureDataset("train.csv")
        >>> train_data.normalize_y()
        >>> test_data = FeatureDataset("test.csv")
        >>> test_data.normalize_y(train_data.y_mean, train_data.y_scale)
        """

        self.targets, self.y_mean, self.y_scale = self._get_normalized(
            self.get_valid_shape_array(self.targets), mean, scale
        )
        self.targets = torch.tensor(self.targets, dtype=torch.float)
        return self

    def __len__(self):
        return len(self.targets)

    def __getitem__(self, idx):
        return self.X[idx], self.targets[idx]



import tarfile
from pathlib import Path
from typing import Union

import pandas as pd

# import subprocess
from PIL import Image
from sklearn.model_selection import StratifiedShuffleSplit
from tqdm import tqdm


class Adience:
    """
    Base class for the Adience dataset.

    Parameters
    ----------
    extract_file_path : Union[str, Path]
        Path to the tar.gz file containing the dataset.
    folds_path : Union[str, Path]
        Path to the folder containing the folds.
    images_path : Union[str, Path]
        Path to the folder containing the images.
    transformed_images_path : Union[str, Path]
        Path to the folder containing the transformed images.
    partition_path : Union[str, Path]
        Path to the folder containing the partitions.
    number_partitions : int, optional
        Number of partitions to create, by default 20.
    ranges : list, optional
        List of age ranges to use, by default [(0, 2), (4, 6), (8, 13),
        (15, 20), (25, 32), (38, 43), (48, 53), (60, 100)].
    test_size : float, optional
        Test size, by default 0.2.
    extract : bool, optional
        Boolean indicating if the tar.gz file should be extracted, by default True.
    transfrom : bool, optional
        Boolean indicating if the images should be transformed and the partitions
        created, by default True.
    """

    def __init__(
        self,
        extract_file_path: Union[str, Path],
        folds_path: Union[str, Path],
        images_path: Union[str, Path],
        transformed_images_path: Union[str, Path],
        partition_path: Union[str, Path],
        number_partitions: int = 20,
        ranges: list = [
            (0, 2),
            (4, 6),
            (8, 13),
            (15, 20),
            (25, 32),
            (38, 43),
            (48, 53),
            (60, 100),
        ],
        test_size: float = 0.2,
        extract: bool = True,
        transfrom: bool = True,
    ) -> None:
        super().__init__()

        self.extract_file_path = Path(extract_file_path)
        self.folds_path = Path(folds_path)
        self.images_path = Path(images_path)
        self.transformed_images_path = Path(transformed_images_path)
        self.partition_path = Path(partition_path)

        self.number_partitions = number_partitions
        self.ranges = ranges
        self.test_size = test_size

        folds = [
            pd.read_csv(self.folds_path / f"fold_{f}_data.txt", sep="\t")
            for f in range(5)
        ]

        if extract:
            self.extract_tar_gz()

        if transfrom:
            self.process_and_split(folds)

    def _check_if_extracted(self) -> bool:
        """
        Check if the tar.gz file has been extracted.
        """
        path = self.extract_file_path.parent
        path = path / "aligned"
        return path.exists()

    def _check_if_transformed(self) -> bool:
        """
        Check if the images have been transformed and the partitions created.
        """
        return self.transformed_images_path.exists() or self.partition_path.exists()

    def extract_tar_gz(self):
        """
        Extract the tar.gz file.
        """
        if self._check_if_extracted():
            print("File already extracted.")
            return

        print("Extracting file...")
        with tarfile.open(self.extract_file_path, "r:gz") as file:
            path = self.extract_file_path.parent
            path.mkdir(exist_ok=True, parents=True)
            file.extractall(path, members=self.track_progress(file))
            file.close()

    def process_and_split(self, folds: list) -> None:
        """
        Process the folds and split the images into partitions.

        Parameters
        ----------
        folds : list
            List of folds.
        """
        if self._check_if_transformed():
            print("File already transformed.")
            return

        fold_dfs = list()
        for f, fold in enumerate(folds):
            fold = fold.assign(age=fold["age"].map(self.assign_range))
            notna = fold["age"].notna()
            n_discarded = (~notna).sum()
            print(
                f"Fold {f}: discarding {n_discarded} entries"
                f" ({(n_discarded / len(fold)) * 100:.1f}%)"
            )
            fold = fold.loc[notna]
            fold = fold.assign(age=fold["age"].astype(int))

            fold_dfs.append(
                pd.DataFrame(
                    dict(
                        path=fold.apply(self.image_path_from_row, axis="columns"),
                        age=fold["age"],
                    )
                )
            )
        df: pd.DataFrame = pd.concat(fold_dfs, ignore_index=True)  # type: ignore

        self.transformed_images_path.mkdir(exist_ok=True)
        print("Resizing images...")
        for row in tqdm(df.itertuples(), total=len(df)):
            dst_image = self.transformed_images_path / row.path
            if dst_image.is_file():
                continue
            src_image = self.images_path / row.path
            dst_image.parent.mkdir(exist_ok=True, parents=True)

            # open the source image
            with Image.open(src_image) as img:
                # calculate the new width that maintains the aspect ratio
                width_percent = 128 / float(img.size[1])
                new_width = int((float(img.size[0]) * float(width_percent)))

                # resize the image using the calculated width and 128 height
                resized_img = img.resize((new_width, 128))

                # save the resized image to the destination path
                resized_img.save(dst_image)

        self.partition_path.mkdir(exist_ok=True)
        for partition in range(self.number_partitions):
            for c in range(len(self.ranges)):
                (self.partition_path / f"{partition}/train/{c}").mkdir(
                    parents=True, exist_ok=True
                )
                (self.partition_path / f"{partition}/test/{c}").mkdir(
                    parents=True, exist_ok=True
                )

        sss = StratifiedShuffleSplit(
            self.number_partitions, test_size=self.test_size, random_state=0
        )
        for partition, (train_index, test_index) in tqdm(
            enumerate(sss.split(df, df["age"]))
        ):
            train_df: pd.DataFrame = df.iloc[train_index]  # type: ignore
            test_df: pd.DataFrame = df.iloc[test_index]  # type: ignore
            for name, partition_df in zip(("train", "test"), (train_df, test_df)):
                for row in tqdm(
                    partition_df.itertuples(),
                    total=len(partition_df),
                    leave=False,
                    desc=name,
                ):
                    image_path = self.transformed_images_path / row.path
                    assert image_path.is_file()
                    new_path = (
                        self.partition_path
                        / f"{partition}/{name}/{row.age}/{image_path.name}"
                    )
                    if not new_path.exists():
                        new_path.symlink_to(image_path.resolve())

    def assign_range(self, age: str):
        """
        Assign an age range to an age.

        Parameters
        ----------
        age : str
            Age to assign a range to.
        """
        age = eval(age)

        if age is None:
            return None

        if age in self.ranges:
            return self.ranges.index(age)

        if isinstance(age, tuple):
            age_minimum, age_maximum = age
            for i, (range_minimum, range_maximum) in enumerate(self.ranges):
                if (age_minimum >= range_minimum) and (age_maximum <= range_maximum):
                    return i
            return None

        if isinstance(age, int):
            for i, (range_minimum, range_maximum) in enumerate(self.ranges):
                if (age >= range_minimum) and (age <= range_maximum):
                    return i
            return None

        return None

    def image_path_from_row(self, row):
        """
        Get the image path from a row.

        Parameters
        ----------
        row : pd.Series
            Row to get the image path from.
        """
        return f'{row["user_id"]}/landmark_aligned_face.{row["face_id"]}.{row["original_image"]}'

    def track_progress(self, file):
        """
        Track the progress of the extraction.

        Parameters
        ----------
        file : tarfile.TarFile
            File to track the progress of.
        """
        for member in tqdm(file, total=len(file.getmembers())):
            # this will be the current file being extracted
            # Go over each member
            yield member



import json
import os
from pathlib import Path
from typing import Callable, Dict, Optional

import numpy as np
from sklearn.metrics import confusion_matrix, recall_score


def minimum_sensitivity(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes the sensitivity by class and returns the lowest value.

    Parameters
    ----------
    y_true : array-like
            Target labels.
    y_pred : array-like
            Predicted probabilities or labels.

    Returns
    -------
    ms : float
            Minimum sensitivity.

    Examples
    --------
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_pred = np.array([0, 1, 0, 1])
    >>> minimum_sensitivity(y_true, y_pred)
    0.5
    """

    sensitivities = recall_score(y_true, y_pred, average=None)
    return np.min(sensitivities)


def accuracy_off1(y_true: np.ndarray, y_pred: np.ndarray, labels=None) -> float:
    """Computes the accuracy of the predictions, allowing errors if they occur in an adjacent class.

    Parameters
    ----------
    y_true : array-like
            Target labels.
    y_pred : array-like
            Predicted probabilities or labels.
    labels : array-like or None
            Labels of the classes. If None, the labels are inferred from the data.

    Returns
    -------
    acc : float
            1-off accuracy.

    Examples
    --------
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_pred = np.array([0, 1, 0, 1])
    >>> accuracy_off1(y_true, y_pred)
    1.0
    """

    if len(y_true.shape) > 1:
        y_true = np.argmax(y_true, axis=1)
    if len(y_pred.shape) > 1:
        y_pred = np.argmax(y_pred, axis=1)
    if labels is None:
        labels = np.unique(y_true)

    conf_mat = confusion_matrix(y_true, y_pred, labels=labels)
    n = conf_mat.shape[0]
    mask = np.eye(n, n) + np.eye(n, n, k=1), +np.eye(n, n, k=-1)
    correct = mask * conf_mat

    return 1.0 * np.sum(correct) / np.sum(conf_mat)


def gmsec(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Geometric mean of the sensitivity of the extreme classes.
    Determines how good the classification performance for the first and the last
    classes is.

    Parameters
    ----------
    y_true : array-like
            Target labels.
    y_pred : array-like
            Predicted probabilities or labels.

    Returns
    -------
    gmec : float
            Geometric mean of the sensitivities of the extreme classes.

    Examples
    --------
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_pred = np.array([0, 1, 0, 1])
    >>> gmec(y_true, y_pred)
    0.5
    """

    sensitivities = recall_score(y_true, y_pred, average=None)
    return np.sqrt(sensitivities[0] * sensitivities[-1])


def amae(y_true: np.ndarray, y_pred: np.ndarray):
    """Computes the average mean absolute error computed independently for each class.

    Parameters
    ----------
    y_true : array-like
            Targets labels with one-hot or integer encoding.
    y_pred : array-like
            Predicted probabilities or labels.

    Returns
    -------
    amae : float
            Average mean absolute error.
    """

    if len(y_true.shape) > 1:
        y_true = np.argmax(y_true, axis=1)
    if len(y_pred.shape) > 1:
        y_pred = np.argmax(y_pred, axis=1)

    cm = confusion_matrix(y_true, y_pred)
    n_class = cm.shape[0]
    costs = np.reshape(np.tile(range(n_class), n_class), (n_class, n_class))
    costs = np.abs(costs - np.transpose(costs))
    non_zero_cm_rows = ~np.all(cm == 0, axis=1)
    cm_ = cm[non_zero_cm_rows]
    errors = costs * cm_
    per_class_maes = np.sum(errors, axis=1) / np.sum(cm_, axis=1).astype("double")
    return np.mean(per_class_maes)


def mmae(y_true: np.ndarray, y_pred: np.ndarray):
    """Computes the maximum mean absolute error computed independently for each class.

    Parameters
    ----------
    y_true : array-like
            Target labels with one-hot or integer encoding.
    y_pred : array-like
            Predicted probabilities or labels.

    Returns
    -------
    mmae : float
            Maximum mean absolute error.
    """

    if len(y_true.shape) > 1:
        y_true = np.argmax(y_true, axis=1)
    if len(y_pred.shape) > 1:
        y_pred = np.argmax(y_pred, axis=1)

    cm = confusion_matrix(y_true, y_pred)
    n_class = cm.shape[0]
    costs = np.reshape(np.tile(range(n_class), n_class), (n_class, n_class))
    costs = np.abs(costs - np.transpose(costs))
    non_zero_cm_rows = ~np.all(cm == 0, axis=1)
    cm_ = cm[non_zero_cm_rows]
    errors = costs * cm_
    per_class_maes = np.sum(errors, axis=1) / np.sum(cm_, axis=1).astype("double")
    return per_class_maes.max()


def write_metrics_dict_to_file(
    metrics: Dict[str, float],
    path_str: str,
    filter_fn: Optional[Callable[[str, float], bool]] = lambda n, v: True,
) -> None:
    """Writes a dictionary of metrics to a tabular file.
    The dictionary is filtered by the filter function.
    The first time that the metrics are saved to the file, the keys are written as the header.
    Subsequent calls append the values to the file.

    Parameters
    ----------
    metrics : Dict[str, float]
            Dictionary of metric names associated with their value.
    path_str : str
            Path to the file that will be saved.
            The directory of the file will be created if it does not exist.
            If the file exists, the metrics will be appended to the file in a new row.
    filter_fn : Optional[Callable[[str, bool], bool]], default=lambda n, v: True
            Function that filters the metrics.
            The function takes the name and the value of the metric and returns ``True`` if the metric should be saved.

    Examples
    --------
    >>> metrics = {'acc': 0.5, 'gmsec': 0.25}
    >>> write_metrics_dict_to_file(metrics, 'results.txt')
    >>> write_metrics_dict_to_file(metrics, 'results.txt')
    >>> with open('results.txt', 'r') as f:
    ...     print(f.read())
    acc	gmsec
    0.5	0.25
    0.5	0.25

    >>> write_metrics_dict_to_file(metrics, 'results.txt', filter_fn=lambda name, value: name == 'acc')
    >>> with open('results.txt', 'r') as f:
    ...     print(f.read())
    acc
    0.5
    0.5
    """

    path = Path(path_str)
    directory = path.parents[0]
    os.makedirs(directory, exist_ok=True)

    if not path.is_file():
        with open(path, "w") as f:
            for k, v in metrics.items():
                if filter_fn(k, v):
                    f.write(f"{k},")
            f.write("\n")

    with open(path, "a") as f:
        for k, v in metrics.items():
            if filter_fn(k, v):
                f.write(f"{v},")
        f.write("\n")


def write_array_to_file(array: np.ndarray, path_str: str, id: str):
    """Writes an array to a json file.
    The array is saved as a dictionary with the key 'id' and the value 'array'.

    Parameters
    ----------
    array : array-like
            Array to be saved.
    path_str : str
            Path to the file that will be saved.
            The directory of the file will be created if it does not exist.
    id : str
            Id of the array.

    Examples
    --------
    >>> array = np.array([0, 1, 2])
    >>> write_array_to_file(array, 'results.json', 'array')
    >>> with open('results.json', 'r') as f:
    ...     print(f.read())
    {"array": [0, 1, 2]}

    >>> array2 = np.array([3, 4, 5])
    >>> write_array_to_file(array, 'results.json', 'array2')
    >>> with open('results.json', 'r') as f:
    ...     print(f.read())
    {"array": [0, 1, 2], "array2": [3, 4, 5]}
    """

    path = Path(path_str)
    directory = path.parents[0]
    os.makedirs(directory, exist_ok=True)

    if path.is_file():
        with open(path, "r") as f:
            data = json.load(f)
    else:
        data = dict()

    data[id] = array.tolist()

    with open(path, "w") as f:
        json.dump(data, f)



import torch
from torch.nn import Module


class StickBreakingLayer(Module):
    """Base class to implement the stick breaking layer from :footcite:t:`liu2020unimodal`.

    Parameters
    ----------
    input_shape: int
        Input shape
    num_classes: int
        Number of classes
    """

    def __init__(self, input_shape: int, num_classes: int) -> None:
        super().__init__()
        self.fcn1 = torch.nn.Linear(input_shape, num_classes)
        self.fcn2 = torch.nn.Sigmoid()

    def get_stick_logits(self, x: torch.Tensor):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns:
        --------
        logits : torch.Tensor
            Logits of the stick breaking layer
        """

        # Clamps all elements in input into the range [ min, max ]. Letting min_value
        # and max_value be min and max, respectively
        x = torch.clamp(x, 0.1, 0.9)
        comp = 1.0 - x

        # cumprod is the cumulative product of the elements of the input tensor in
        # the given dimension dim.
        cumprod = torch.cumprod(comp, axis=1)
        logits = torch.log(x * cumprod)
        return logits

    def forward(self, x) -> torch.Tensor:
        """
        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns:
        --------
        logits : torch.Tensor
            Logits of the stick breaking layer
        """

        x = self.fcn1(x)
        x = self.fcn2(x)
        logits = self.get_stick_logits(x)
        return logits



import numpy as np
from scipy.special import softmax


def get_exponential_soft_labels(J, p=1.0, tau=1.0):
    """Get soft labels from exponential distribution for ``J`` classes or splits as
    described in :footcite:t:`liu2020unimodal` and :footcite:t:`vargas2023exponential`.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the exponential
    function for the interval boundaries. The probability for the first interval is
    computed as the value of the exponential function for the first interval boundary.
    Then, a softmax function is applied to each row of the resulting matrix to obtain
    valid probabilities.

    The aforementioned exponential function is defined as follows:
    :math:`f(x; k, p, \\tau) = \\frac{-|x - k|^p}{\\tau}` where :math:`k` is given by the order of the
    true class for which the probability is computed, :math:`p` is the exponent
    parameter and :math:`\\tau` is the scaling parameter.

    Parameters
    ----------
    J : int
            Number of classes.
    p : float, default=1.0
            Exponent parameter :math:`p`.
    tau : float, default=1.0
            Scaling parameter :math:`\\tau`.

    Raises
    ------
    ValueError
            If ``J`` is not a positive integer greater than 1.

    Returns
    -------
    probs : 2d array-like of shape (J, J)
            Matrix of probabilities where each row represents the true class
            and each column the probability for class j.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_exponential_soft_labels
    >>> get_exponential_soft_labels(5)
    array([[0.63640865, 0.23412166, 0.08612854, 0.03168492, 0.01165623],
    [0.19151597, 0.52059439, 0.19151597, 0.07045479, 0.02591887],
    [0.06745081, 0.1833503 , 0.49839779, 0.1833503 , 0.06745081],
    [0.02591887, 0.07045479, 0.19151597, 0.52059439, 0.19151597],
    [0.01165623, 0.03168492, 0.08612854, 0.23412166, 0.63640865]])
    """

    if J < 2 or not isinstance(J, int):
        raise ValueError(f"{J=} must be a positive integer greater than 1")

    probs = []

    for true_class in range(0, J):
        probs.append(-(np.abs(np.arange(0, J) - true_class) ** p) / tau)

    return softmax(np.array(probs), axis=1)



from torch import Tensor
import numpy as np
import torch

import torch.nn as nn
from collections import OrderedDict, namedtuple


class OBDECOCModel(nn.Module):
    """Ordinal Binary Decomposition (OBD) wrapper model from :footcite:t:`barbero2023error`.
    It transforms the output of the provided base classifier into
    ``num_classes - 1`` outputs bounded between 0 and 1, representing
    the class threshold probabilities :math:`P(y > q)`.

    Parameters
    ----------
    num_classes : int
        Number of classes.
    base_classifier: nn.Module
        Base classifier that will be wrapped.
    base_n_outputs:
        Number of outputs of the base classifier. The models implemented
        in ``torchvision`` have 1000 as the default
    """

    num_classes: int

    def __init__(
        self, num_classes: int, base_classifier: nn.Module, base_n_outputs: int
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.base_classifier = base_classifier
        self.obd_output = nn.Sequential(
            OrderedDict(
                [
                    ("penultimate_activation", nn.ReLU()),
                    ("last_linear", nn.Linear(base_n_outputs, num_classes - 1)),
                    ("last_activation", nn.Sigmoid()),
                ]
            )
        )
        self.transformer = ECOCOutputTransformer(num_classes)

    def forward(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : torch.Tensor
            Input to the model

        Returns
        -------
        threshold_probas : torch.Tensor
            Predicted threshold probabilities
        """
        x = self.base_classifier(x)
        x = self.obd_output(x)
        return x

    def predict_from_inputs(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input to the model

        Returns
        -------
        An object with the following attributes.

        scores : Tensor
            The negative distance to each class ideal vector, to use
            as class scores.
        probas : Tensor
            The predicted probability of belonging to each class :math:`P(y = q)`.
        labels : Tensor
            The predicted integer label according to the ECOC assignment
            scheme.
        """
        raw_output = self(x)
        return PredictOutput(
            self.transformer.scores(raw_output),
            self.transformer.probas(raw_output),
            self.transformer.labels(raw_output),
        )


PredictOutput = namedtuple("PredictOutput", ["scores", "probas", "label"])


class ECOCOutputTransformer(nn.Module):
    """A transformer for the output of the OBD model in order
    to apply the ECOC scheme.

    Parameters
    ----------
    num_classes : int
        Number of classes.
    """

    target_class: Tensor

    def __init__(self, num_classes: int) -> None:
        super().__init__()
        self.num_classes = num_classes
        target_class = np.ones((num_classes, num_classes - 1), dtype=np.float32)
        target_class[np.triu_indices(num_classes, 0, num_classes - 1)] = 0.0
        target_class = torch.tensor(target_class, dtype=torch.float32)
        self.register_buffer("target_class", target_class)

    def probas(self, output):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input to the model

        Returns
        -------
        probas : Tensor
            The predicted probability of belonging to each class :math:`P(y = q)`.
        """
        return torch.softmax(self.scores(output), dim=1)

    def scores(self, output):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input to the model

        Returns
        -------
        scores : Tensor
            The negative distance to each class ideal vector, to use
            as class scores.
        """
        return -torch.cdist(output, self.target_class)

    def labels(self, output):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input to the model

        Returns
        -------
        labels : Tensor
            The predicted integer label according to the ECOC assignment
            scheme.
        """
        scores = self.scores(output)
        return scores.argmax(dim=1)



from typing import Optional

import torch
from torch import Tensor
from torch.nn import CrossEntropyLoss
from torch.nn.functional import one_hot


class CustomTargetsCrossEntropyLoss(torch.nn.Module):
    """Base class to implement a unimodal regularised cross entropy loss.

    Parameters
    ----------
    cls_probs : Tensor
        The class probabilities tensor.
    eta : float, default=1.0
        Parameter that controls the influence of the regularisation.
    weight : Optional[Tensor], default=None
        A manual rescaling weight given to each class. If given, has to be a Tensor
        of size `C`. Otherwise, it is treated as if having all ones.
    size_average : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged over
        each loss element in the batch. Note that for some losses, there are
        multiple elements per sample. If the field :attr:`size_average` is set to
        ``False``, the losses are instead summed for each minibatch. Ignored when
        reduce is ``False``. Default: ``True``
    ignore_index : int, default=-100
        Specifies a target value that is ignored and does not contribute to the
        input gradient. When :attr:`size_average` is ``True``, the loss is averaged
        over non-ignored targets.
    reduce : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged or
        summed over observations for each minibatch depending on :attr:`size_average`.
        When :attr:`reduce` is ``False``, returns a loss per batch element instead
        and ignores :attr:`size_average`. Default: ``True``
    reduction : str, default='mean'
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` |
        ``'sum'``. ``'none'``: no reduction will be applied, ``'mean'``: the sum of
        the output will be divided by the number of elements in the output,
        ``'sum'``: the output will be summed. Note: :attr:`size_average` and
        :attr:`reduce` are in the process of being deprecated, and in the meantime,
        specifying either of those two args will override :attr:`reduction`.
        Default: ``'mean'``
    label_smoothing : float, default=0.0
        Controls the amount of label smoothing for the loss. Zero means no smoothing.
        Default: ``0.0``
    """

    def __init__(
        self,
        cls_probs: Tensor,
        eta: float = 1.0,
        weight: Optional[Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
        label_smoothing: float = 0.0,
    ):
        super().__init__()

        self.num_classes = cls_probs.size(0)
        self.eta = eta

        self.ce_loss = CrossEntropyLoss(
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
            label_smoothing=label_smoothing,
        )

        # Default class probs initialized to ones
        self.register_buffer("cls_probs", cls_probs.float())

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """
        Method that is called to compute the loss.

        Parameters
        ----------
        input : Tensor
            The input tensor.
        target : Tensor
            The target tensor.

        Returns
        -------
        loss: Tensor
            The computed loss.
        """

        y_prob = self.get_buffer("cls_probs")[target]
        target_oh = one_hot(target, self.num_classes)

        y_true = (1.0 - self.eta) * target_oh + self.eta * y_prob

        return self.ce_loss(input, y_true)



from .beta_distribution import get_beta_soft_labels
from .binomial_distribution import get_binomial_soft_labels
from .exponential_distribution import get_exponential_soft_labels
from .general_triangular_distribution import (
    get_general_triangular_params,
    get_general_triangular_soft_labels,
)
from .poisson_distribution import get_poisson_soft_labels
from .triangular_distribution import get_triangular_soft_labels

__all__ = [
    "get_beta_soft_labels",
    "get_exponential_soft_labels",
    "get_binomial_soft_labels",
    "get_poisson_soft_labels",
    "get_triangular_soft_labels",
    "get_general_triangular_params",
    "get_general_triangular_soft_labels",
]



import warnings
from typing import Literal

import torch
from torch.nn import Module


class CLM(Module):
    """
    Implementation of the cumulative link models from :footcite:t:`vargas2020clm` as a
    torch layer. Different link functions can be used, including logit, probit
    and cloglog.

    Parameters
    ----------
    num_classes : int
        The number of classes.
    link_function : str
        The link function to use. Can be ``'logit'``, ``'probit'`` or ``'cloglog'``.
    min_distance : float, default=0.0
        The minimum distance between thresholds
    clip_warning : bool, default=True
        Whether to print the clipping value warning or not.

    Attributes
    ----------
    num_classes : int
        The number of classes.
    link_function : str
        The link function to use. Can be ``'logit'``, ``'probit'`` or ``'cloglog'``.
    min_distance : float
        The minimum distance between thresholds
    clip_warning : bool
        Whether to print the clipping value warning or not.
    dist_ : torch.distributions.Normal
        The normal (0,1) distribution used to compute the probit link function.
    thresholds_b_ : torch.nn.Parameter
        The torch parameter for the first threshold.
    thresholds_a_ : torch.nn.Parameter
        The torch parameter for the alphas of the thresholds.
    clip_warning_shown_ : bool
        Whether the clipping warning has been shown or not.


    Example
    ---------
    >>> import torch
    >>> from dlordinal.output_layers import CLM
    >>> inp = torch.randn(10, 5)
    >>> fc = torch.nn.Linear(5, 1)
    >>> clm = CLM(5, "logit")
    >>> output = clm(fc(inp))
    >>> print(output)
    tensor([[0.7944, 0.1187, 0.0531, 0.0211, 0.0127],
            [0.4017, 0.2443, 0.1862, 0.0987, 0.0690],
            [0.4619, 0.2381, 0.1638, 0.0814, 0.0548],
            [0.4636, 0.2378, 0.1632, 0.0809, 0.0545],
            [0.4330, 0.2419, 0.1746, 0.0893, 0.0612],
            [0.5006, 0.2309, 0.1495, 0.0716, 0.0473],
            [0.6011, 0.2027, 0.1138, 0.0504, 0.0320],
            [0.5995, 0.2032, 0.1144, 0.0507, 0.0322],
            [0.4014, 0.2443, 0.1863, 0.0988, 0.0691],
            [0.6922, 0.1672, 0.0838, 0.0351, 0.0217]], grad_fn=<CopySlices>)

    """

    def __init__(
        self,
        num_classes: int,
        link_function: Literal["logit", "probit", "cloglog"],
        min_distance: int = 0.0,
        clip_warning: bool = True,
        **kwargs,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.link_function = link_function
        self.min_distance = min_distance
        self.clip_warning = clip_warning
        self.dist_ = torch.distributions.Normal(0, 1)

        self.thresholds_b_ = torch.nn.Parameter(
            data=torch.Tensor([0]), requires_grad=True
        )
        self.thresholds_a_ = torch.nn.Parameter(
            data=torch.Tensor([1.0 for _ in range(self.num_classes - 2)]),
            requires_grad=True,
        )

        self.clip_warning_shown_ = False

    def _convert_thresholds(self, b, a, min_distance):
        a = a**2
        a = a + min_distance
        thresholds_param = torch.cat((b, a), dim=0)
        th = torch.sum(
            torch.tril(
                torch.ones(
                    (self.num_classes - 1, self.num_classes - 1), device=a.device
                ),
                diagonal=0,
            )
            * torch.reshape(
                torch.tile(thresholds_param, (self.num_classes - 1,)).to(a.device),
                shape=(self.num_classes - 1, self.num_classes - 1),
            ),
            dim=(1,),
        )
        return th

    def _compute_z3(self, projected: torch.Tensor, thresholds: torch.Tensor):
        m = projected.shape[0]
        a = torch.reshape(torch.tile(thresholds, (m,)), shape=(m, -1))
        b = torch.transpose(
            torch.reshape(
                torch.tile(projected, (self.num_classes - 1,)), shape=(-1, m)
            ),
            0,
            1,
        )

        z3 = a - b
        if torch.any(z3 > 10) or torch.any(z3 < -10):
            if self.clip_warning and not self.clip_warning_shown_:
                warnings.warn(
                    f"The output value of the CLM layer (max: {z3.abs().max()}) is out "
                    "of the range [-10, 10]. Clipping value prior to applying the "
                    "link function for numerical stability."
                )
            z3 = torch.clip(a - b, -10, 10)
            self.clip_warning_shown_ = True

        return z3

    def _apply_link_function(self, z3):
        if self.link_function == "probit":
            a3T = self.dist_.cdf(z3)
        elif self.link_function == "cloglog":
            a3T = 1 - torch.exp(-torch.exp(z3))
        else:
            a3T = 1.0 / (1.0 + torch.exp(-z3))

        return a3T

    def _clm(self, projected: torch.Tensor, thresholds: torch.Tensor):
        projected = torch.reshape(projected, shape=(-1,))

        m = projected.shape[0]
        z3 = self._compute_z3(projected, thresholds)
        a3T = self._apply_link_function(z3)

        ones = torch.ones((m, 1), device=projected.device)
        a3 = torch.cat((a3T, ones), dim=1)
        a3[:, 1:] = a3[:, 1:] - a3[:, 0:-1]

        return a3

    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor.

        Returns
        -------
        output: Tensor
            The output tensor.
        """

        thresholds = self._convert_thresholds(
            self.thresholds_b_, self.thresholds_a_, self.min_distance
        )

        return self._clm(x, thresholds)



from typing import Optional

import torch
import torch.nn as nn


class WKLoss(nn.Module):
    """
    Implements Weighted Kappa Loss. Weighted Kappa Loss was introduced by :footcite:t:`deLaTorre2018kappa`.
    Weighted Kappa is widely used in Ordinal Classification Problems. Its
    value lies in :math:`[-\\infty, \\log 2]`, where :math:`\\log 2` means the random prediction

    Parameters
    ----------
    num_classes : int
        Number of unique classes in your dataset.
    penalization_type : str, default='quadratic'
        Weighting to be considered for calculating kappa
        statistics. A valid value is one of ``['linear', 'quadratic']``.
        Defaults to 'quadratic'.
    epsilon : float, default=1e-10
        Increment to avoid log zero,
        so the loss will be :math:`\\log(1 - k + \\epsilon)`, where :math:`k` lies
        in :math:`[-1, 1]`. Defaults to ``1e-10``.
    """

    def __init__(
        self,
        num_classes: int,
        penalization_type: str = "quadratic",
        weight: Optional[torch.Tensor] = None,
        epsilon: Optional[float] = 1e-10,
    ):
        super(WKLoss, self).__init__()
        self.num_classes = num_classes
        if penalization_type == "quadratic":
            self.y_pow = 2
        if penalization_type == "linear":
            self.y_pow = 1

        self.epsilon = epsilon
        self.weight = weight

    def forward(self, y_pred, y_true):
        num_classes = self.num_classes
        y = torch.eye(num_classes)
        y_true = y[y_true]

        y_true = y_true.float()

        repeat_op = (
            torch.Tensor(list(range(num_classes))).unsqueeze(1).repeat((1, num_classes))
        )
        repeat_op_sq = torch.square((repeat_op - repeat_op.T))
        weights = repeat_op_sq / ((num_classes - 1) ** 2)

        # Apply class weight
        if self.weight is not None:
            # Repeat weight num_classes times in columns
            tiled_weight = self.weight.repeat((num_classes, 1))
            weights *= tiled_weight

        pred_ = y_pred**self.y_pow
        pred_norm = pred_ / (self.epsilon + torch.reshape(torch.sum(pred_, 1), [-1, 1]))

        hist_rater_a = torch.sum(pred_norm, 0)
        hist_rater_b = torch.sum(y_true, 0)

        conf_mat = torch.matmul(pred_norm.T, y_true)

        bsize = y_pred.size(0)
        nom = torch.sum(weights * conf_mat)
        expected_probs = torch.matmul(
            torch.reshape(hist_rater_a, [num_classes, 1]),
            torch.reshape(hist_rater_b, [1, num_classes]),
        )
        denom = torch.sum(weights * expected_probs / bsize)

        return nom / (denom + self.epsilon)



from typing import Optional

import torch
from torch import Tensor

from ..soft_labelling import get_triangular_soft_labels
from .custom_targets_loss import CustomTargetsCrossEntropyLoss


class TriangularCrossEntropyLoss(CustomTargetsCrossEntropyLoss):
    """Triangular unimodal regularised cross entropy loss from :footcite:t:`vargas2023softlabelling`.

    Parameters
    ----------
    num_classes : int
        Number of classes.
    alpha2 : float, default=0.05
        Parameter that controls the probability deposited in adjacent classes.
    eta : float, default=1.0
        Parameter that controls the influence of the regularisation.
    weight : Optional[Tensor], default=None
        A manual rescaling weight given to each class. If given, has to be a Tensor
        of size `C`. Otherwise, it is treated as if having all ones.
    size_average : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged over
        each loss element in the batch. Note that for some losses, there are
        multiple elements per sample. If the field :attr:`size_average` is set to
        ``False``, the losses are instead summed for each minibatch. Ignored when
        reduce is ``False``. Default: ``True``
    ignore_index : int, default=-100
        Specifies a target value that is ignored and does not contribute to the
        input gradient. When :attr:`size_average` is ``True``, the loss is averaged
        over non-ignored targets.
    reduce : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged or
        summed over observations for each minibatch depending on :attr:`size_average`.
        When :attr:`reduce` is ``False``, returns a loss per batch element instead
        and ignores :attr:`size_average`. Default: ``True``
    reduction : str, default='mean'
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` |
        ``'sum'``. ``'none'``: no reduction will be applied, ``'mean'``: the sum of
        the output will be divided by the number of elements in the output,
        ``'sum'``: the output will be summed. Note: :attr:`size_average` and
        :attr:`reduce` are in the process of being deprecated, and in the meantime,
        specifying either of those two args will override :attr:`reduction`.
        Default: ``'mean'``
    """

    def __init__(
        self,
        num_classes: int,
        alpha2: float = 0.05,
        eta: float = 1.0,
        weight: Optional[Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
    ):
        # Precompute class probabilities for each label
        cls_probs = torch.tensor(get_triangular_soft_labels(num_classes, alpha2))
        super().__init__(
            cls_probs=cls_probs,
            eta=eta,
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
            label_smoothing=0.0,
        )



import math

import numpy as np
from sympy import Eq, nsolve, solve, symbols
from sympy.core.add import Add

from .utils import get_intervals, triangular_cdf


def get_general_triangular_params(J: int, alphas: np.ndarray, verbose: int = 0):
    """
    Get the parameters (:math:`a`, :math:`b` and :math:`c`) for the
    general triangular distribution. The parameters are computed using
    the :math:`\\boldsymbol{\\alpha}` values and the number of classes
    :math:`J`. The :math:`\\boldsymbol{\\alpha}` vector contains two :math:`\\alpha`
    values for each class or split. :math:`\\alpha_0` should always be 0, given that
    the error on the left side of the first class is always 0. In the same way, the
    :math:`\\alpha_{2J}` value should always be 0, given that the error on the right
    side of the last class is always 0. The :math:`\\alpha` values for the other
    classes should be between 0 and 1.

    The parameters :math:`a`, :math:`b` and :math:`c` for class :math:`j` are computed
    as described in :footcite:t:`vargas2023gentri`.

    Parameters
    ----------
    J : int
        Number of classes or splits.
    alphas : np.ndarray
        Array that represents the error on the left and right side of each class.
        It is the :math:`\\boldsymbol{\\alpha}` vector described
        in :footcite:t:`vargas2023gentri`.
    verbose : int, optional
        Verbosity level, by default 0.

    Raises
    ------
    ValueError
        If the number of classes :math:`J` is less than 2.
        If the :math:`\\boldsymbol{\\alpha}` vector is not a numpy array of shape
        :math:`(2J,)`.

    Returns
    -------
    list
        List of dictionaries with the parameters for each class.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_general_triangular_params
    >>> get_general_triangular_params(5, [0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.1, 0])
    [{'alpha2j_1': 0, 'alpha2j': 0.05, 'a': 0, 'b': 0.25760143110525874, 'c': 0}, {'alpha2j_1': 0.05, 'alpha2j': 0.05, 'a': 0.153752470442574, 'b': 0.446247529557426, 'c': 0.3}, {'alpha2j_1': 0.05, 'alpha2j': 0.05, 'a': 0.353752470442574, 'b': 0.646247529557426, 'c': 0.5}, {'alpha2j_1': 0.05, 'alpha2j': 0.1, 'a': 0.550779686438060, 'b': 0.875486049708105, 'c': 0.7}, {'alpha2j_1': 0.0, 'alpha2j': 0, 'a': 0.8, 'b': 1, 'c': 1}]

    """

    alphas = np.array(alphas)
    if not isinstance(alphas, np.ndarray) or alphas.shape != (2 * J,):
        raise ValueError(
            f"alphas must be a numpy array of shape (2 * n,), but got {alphas.shape}"
        )

    if J < 2:
        raise ValueError(f"J must be greater than 1, but got {J}")

    def abc1(J, alpha):
        a = 0
        b = 1.0 / ((1.0 - math.sqrt(alpha)) * J)
        c = 0
        return a, b, c

    def abcJ(J, alpha):
        a = 1.0 - 1.0 / (J * (1.0 - math.sqrt(alpha)))
        b = 1
        c = 1
        return a, b, c

    def abcj(J, j, alpha2j_1, alpha2j):
        # Calculation of the terms of the system of equations
        # b = ...
        numa2 = 2 * J**2 - 2 * J**2 * alpha2j_1
        numa1 = 2 * j * J * alpha2j_1 - J * alpha2j_1 - 4 * J * j + 4 * J
        numa0 = 2 * j**2 - 4 * j + 2
        dena0 = 2 * j * J * alpha2j_1 - J * alpha2j_1
        dena1 = -(2 * J**2 * alpha2j_1)

        # a = ...
        numb2 = 2 * J**2 * alpha2j - 2 * J**2
        numb1 = 4 * J * j - 2 * J * j * alpha2j + J * alpha2j
        numb0 = -(2 * j**2)
        denb1 = 2 * J**2 * alpha2j
        denb0 = -2 * J * j * alpha2j + J * alpha2j

        a, b = symbols("a, b")
        c = (2 * j - 1) / (2 * J)
        eq1 = Eq((numa2 * a**2 + numa1 * a + numa0) / (dena0 + dena1 * a), b)
        eq2 = Eq((numb2 * b**2 + numb1 * b + numb0) / (denb1 * b + denb0), a)

        try:
            nsol = nsolve([eq1, eq2], [a, b], [1, 1])

            if nsol[0] < (j - 1) / J and nsol[1] > j / J:
                return nsol[0], nsol[1], c
        except ValueError:
            pass

        try:
            sol = solve([eq1, eq2], [a, b])
        except ValueError:
            raise ValueError(
                f"Unsatisfiable alpha values {alphas}: could not solve for the"
                " triangular distribution parameters"
            )

        soln = [tuple(v.evalf() for v in s) for s in sol]

        for s in soln:
            s_a = s[0]
            s_b = s[1]

            if isinstance(s_a, Add):
                s_a = s_a.args[0]
            if isinstance(s_b, Add):
                s_b = s_b.args[0]

            if s_a < (j - 1) / J and s_b > j / J:
                return s_a, s_b, c

        raise ValueError(f"Could not find solution for {j=}, {alpha2j_1=}, {alpha2j=}")

    if verbose >= 3:
        print(f"{abc1(J,alphas[2])=}, {abcJ(J,alphas[2*J-1])=}")
        for i in range(2, J):
            print(f"{i=}  {abcj(J,i,alphas[2*i-1],alphas[2*i])}")

    params = []

    # Compute params for each interval (class)
    for j in range(1, J + 1):
        j_params = {}
        if j == 1:
            a, b, c = abc1(J, alphas[2])
            j_params["alpha2j_1"] = 0
            j_params["alpha2j"] = alphas[2]
        elif j == J:
            a, b, c = abcJ(J, alphas[2 * J - 1])
            j_params["alpha2j_1"] = alphas[2 * J - 1]
            j_params["alpha2j"] = 0
        else:
            a, b, c = abcj(J, j, alphas[2 * j - 1], alphas[2 * j])
            j_params["alpha2j_1"] = alphas[2 * j - 1]
            j_params["alpha2j"] = alphas[2 * j]

        j_params["a"] = a
        j_params["b"] = b
        j_params["c"] = c
        params.append(j_params)

    return params


def get_general_triangular_soft_labels(J: int, alphas: np.ndarray, verbose: int = 0):
    """Get soft labels using triangular distributions for ``J`` classes or splits.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the triangular
    distribution function for the interval boundaries. The probability for the first
    interval is computed as the value of the triangular distribution function for the
    first interval boundary.

    The triangular distribution function is denoted as :math:`\\text{f}(x, a, b, c)`.
    The parameters :math:`a`, :math:`b` and :math:`c` for class :math:`j` are computed
    as described in :footcite:t:`vargas2023gentri`.

    Parameters
    ----------
    J : int
        Number of classes.
    alphas : np.ndarray
        Array of alphas.
    verbose : int, optional
        Verbosity level, by default 0.

    Raises
    ------
    ValueError
        If ``J`` is not a positive integer greater than 1.
        If ``alphas`` is not a numpy array of shape :math:`(2J,)`.

    Returns
    -------
    probs : 2d array-like of shape (J, J)
        Matrix of probabilities where each row represents the true class
        and each column the probability for class j.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_general_triangular_soft_labels
    >>> get_general_triangular_soft_labels(
    ...     5,
    ...     [0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.1, 0]
    ... )
    array([[0.95, 0.05, 0.  , 0.  , 0.  ],
           [0.05, 0.9 , 0.05, 0.  , 0.  ],
           [0.  , 0.05, 0.9 , 0.05, 0.  ],
           [0.  , 0.  , 0.05, 0.85, 0.1 ],
           [0.  , 0.  , 0.  , 0.  , 1.  ]], dtype=float32)
    """

    if J < 2:
        raise ValueError(f"J must be greater than 1, but got {J}")

    if isinstance(alphas, list):
        alphas = np.array(alphas)

    if not isinstance(alphas, np.ndarray) or alphas.shape != (2 * J,):
        raise ValueError(
            "alphas must be a numpy array or list of shape (2 * n,),"
            " but got {alphas.shape}"
        )

    intervals = get_intervals(J)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    params = get_general_triangular_params(J, alphas, verbose)

    for param in params:
        a = param["a"]
        b = param["b"]
        c = param["c"]
        j_probs = []

        for interval in intervals:
            j_probs.append(
                triangular_cdf(interval[1], a, b, c)
                - triangular_cdf(interval[0], a, b, c)
            )
            if verbose >= 2:
                print(f"\tinterval: {interval}, prob={j_probs[-1]}")

        probs.append(j_probs)

    return np.array(probs, dtype=np.float32)



import re
import shutil
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from skimage.io import imread, imsave
from skimage.transform import resize
from skimage.util import img_as_ubyte
from sklearn.model_selection import train_test_split
from torchvision.datasets.utils import download_and_extract_archive
from torchvision.datasets.vision import VisionDataset
from tqdm import tqdm


class FGNet(VisionDataset):
    """
    Base class for FGNet dataset.

    Parameters
    ----------
    root : str or Path
        Root directory of dataset
    download : bool, optional
        If True, downloads the dataset from the internet and puts it in root directory.
        If dataset is already downloaded, it is not downloaded again.
    process_data : bool, optional
        If True, processes the dataset and puts it in root directory.
        If dataset is already processed, it is not processed again.
    target_size : tuple, optional
        Size of the images after resizing.
    categories : list, optional
        List of categories to be used.
    test_size : float, optional
        Size of the test set.
    validation_size : float, optional
        Size of the validation set.
    """

    def __init__(
        self,
        root: Union[str, Path],
        download: bool = False,
        process_data: bool = True,
        target_size: tuple = (128, 128),
        categories: list = [3, 11, 16, 24, 40],
        test_size: float = 0.2,
        validation_size: float = 0.15,
    ) -> None:
        super(FGNet, self).__init__(root)

        self.root = Path(self.root)
        self.root.parent.mkdir(parents=True, exist_ok=True)
        self.target_size = target_size
        self.categories = categories
        self.test_size = test_size
        self.validation_size = validation_size

        original_path = self.root / "FGNET/images"
        processed_path = self.root / "FGNET/data_processed"

        original_csv_path = self.root / "FGNET/data_processed/fgnet.csv"
        train_csv_path = self.root / "FGNET/data_processed/train.csv"
        test_csv_path = self.root / "FGNET/data_processed/test.csv"

        original_images_path = self.root / "FGNET/data_processed"
        train_images_path = self.root / "FGNET/train"
        test_images_path = self.root / "FGNET/test"

        if download:
            self.download()
        if not self._check_integrity_download():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to"
                " download it"
            )

        if process_data:
            self.process(original_path, processed_path)
            self.split(
                original_csv_path,
                train_csv_path,
                test_csv_path,
                original_images_path,
                train_images_path,
                test_images_path,
            )

    def __str__(self) -> str:
        return "FGNet"

    def download(self) -> None:
        """
        Download the FGNet dataset and extract it.
        """
        if self._check_integrity_download():
            print("Files already downloaded and verified")
            return

        download_and_extract_archive(
            "http://yanweifu.github.io/FG_NET_data/FGNET.zip",
            self.root,
            filename="fgnet.zip",
            md5="1206978cac3626321b84c22b24cc8d19",
        )

    def process(self, original_path, processed_path):
        """
        Process the FGNet dataset and save it in the processed_path.

        Parameters
        ----------
        original_path : Path
            Path to the original dataset.
        processed_path : Path
            Path to save the processed dataset.
        """
        if self._check_integrity_process():
            print("Files already processed and verified")
            return

        data = self.load_data(original_path)
        df = pd.DataFrame(data, columns=["path", "category"])
        processed_path.mkdir(parents=True, exist_ok=True)
        df.to_csv(processed_path / "fgnet.csv", index=False)
        self.process_images_from_df(df, original_path, processed_path)
        return df

    def split(
        self,
        original_csv_path: Path,
        train_csv_path: Path,
        test_csv_path: Path,
        original_images_path: Path,
        train_images_path: Path,
        test_images_path: Path,
    ):
        """
        Split the FGNet dataset into train and test sets.

        Parameters
        ----------
        original_csv_path : Path
            Path to the original csv file.
        train_csv_path : Path
            Path to save the train csv file.
        test_csv_path : Path
            Path to save the test csv file.
        original_images_path : Path
            Path to the original images.
        train_images_path : Path
            Path to save the train images.
        test_images_path : Path
            Path to save the test images.
        """
        if self._check_integrity_split():
            print("Files already split and verified")
            return

        train, test = self.split_dataframe(
            original_csv_path, train_images_path, original_images_path, test_images_path
        )

        test.to_csv(test_csv_path, index=False)
        train.to_csv(train_csv_path, index=False)

    def _check_integrity_download(self) -> bool:
        """
        Check if the FGNet dataset is downloaded and extracted.
        """
        return (self.root / "FGNET").exists()

    def _check_integrity_process_split(self) -> bool:
        """
        Check if the FGNet dataset is processed and split.
        """
        return (
            (self.root / "FGNET/data_processed").exists()
            and (self.root / "FGNET/trainval").exists()
            and (self.root / "FGNET/test").exists()
        )

    def _check_integrity_process(self) -> bool:
        """
        Check if the FGNet dataset is processed.
        """
        return (self.root / "FGNET/data_processed").exists()

    def _check_integrity_split(self) -> bool:
        """
        Check if the FGNet dataset is split.
        """
        return (self.root / "FGNET/train").exists() and (
            self.root / "FGNET/test"
        ).exists()

    def get_age_from_filename(self, filename):
        """
        Get the age from the filename.

        Parameters
        ----------
        filename : str
            Filename of the image.
        """
        m = re.match("[0-9]+A([0-9]+).*", filename)
        return int(m.groups()[0])

    def find_category(self, real_age):
        """
        Find the category of the real age.

        Parameters
        ----------
        real_age : int
            Real age of the image.
        """
        for i, age in enumerate(self.categories):
            if real_age < age:
                return i
        return len(self.categories)

    def load_data(self, original_path: Path):
        """
        Load the data from the original_path.

        Parameters
        ----------
        original_path : Path
            Path to the original dataset.
        """
        data = []
        for img in original_path.iterdir():
            age = self.get_age_from_filename(img.name)
            category = self.find_category(age)
            data.append([img.name, category])

        return data

    def process_images_from_df(
        self, df: pd.DataFrame, original_path: Path, processed_path: Path
    ):
        """
        Process the images from the dataframe.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with the images.
        original_path : Path
            Path to the original dataset.
        processed_path : Path
            Path to save the processed dataset.
        """
        for idx, row in tqdm(
            df.iterrows(), total=df.shape[0], desc="Processing images", unit="image"
        ):
            path = original_path / Path(row["path"])
            processed_path_images = processed_path / Path(row["path"])
            img = imread(path)
            img = img_as_ubyte(resize(img, self.target_size, anti_aliasing=True))

            processed_path_images.parent.mkdir(parents=True, exist_ok=True)
            imsave(processed_path_images, img, check_contrast=False)

    def split_dataframe(
        self,
        csv_path: Path,
        train_images_path: Path,
        original_images_path: Path,
        test_images_path: Path,
    ):
        """
        Split the dataframe into train and test sets.

        Parameters
        ----------
        csv_path : Path
            Path to the csv file.
        train_images_path : Path
            Path to save the train images.
        original_images_path : Path
            Path to the original images.
        test_images_path : Path
            Path to save the test images.
        """
        df = pd.read_csv(csv_path)
        x = np.array(df["path"])
        y = np.array(df["category"])
        x_train, x_test, y_train, y_test = train_test_split(
            x, y, test_size=self.test_size, random_state=1
        )

        for path, label in zip(x_train, y_train):
            train_path = train_images_path / str(label)
            train_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(original_images_path / path, train_path / path)

        for path, label in zip(x_test, y_test):
            test_path = test_images_path / str(label)
            test_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(original_images_path / path, test_path / path)

        x_train, x_val, y_train, y_val = train_test_split(
            x_train, y_train, test_size=self.validation_size, random_state=1
        )

        train = np.hstack((x_train[:, np.newaxis], y_train[:, np.newaxis]))
        val = np.hstack((x_val[:, np.newaxis], y_val[:, np.newaxis]))
        test = np.hstack((x_test[:, np.newaxis], y_test[:, np.newaxis]))
        trainval = np.vstack((train, val))

        test_df = pd.DataFrame(data=test, columns=["path", "category"])
        train_df = pd.DataFrame(data=trainval, columns=["path", "category"])

        return train_df, test_df



import math

import numpy as np

from .utils import get_intervals, triangular_cdf


def get_triangular_soft_labels(J: int, alpha2: float = 0.01, verbose: int = 0):
    """
    Get soft labels using triangular distributions for ``J`` classes or splits using
    the approach described in :footcite:t:`vargas2023softlabelling`.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the triangular
    distribution function for the interval boundaries. The probability for the first
    interval is computed as the value of the triangular distribution function for the
    first interval boundary.

    The triangular distribution function is denoted as :math:`\\text{p}(x, a, b, c)`
    where :math:`a`, :math:`b` and :math:`c` are the parameters of the distribution,
    and are determined by the number of classes :math:`J` and the value of the
    :math:`\\alpha_2` parameter. The value of :math:`\\alpha_2` represents the
    probability that is assigned to the adjacent classes of the target class. The
    parameters :math:`a`, :math:`b` and :math:`c` for class :math:`j` are computed
    as follows:

    .. math::
        a_j = \\begin{cases}
            0 & \\text{if } j = 1 \\\\
            \\frac{2j - 2 - 4j\\alpha_2 + 2\\alpha_2 \\pm
              \\sqrt{2\\alpha_2}}{2n(1 - 2\\alpha_2)} & \\text{if } 1 < j < J\\\\
            1 + \\frac{1}{\\pm J (\\sqrt{\\alpha_3} - 1)} & \\text{if } j = J
        \\end{cases}

    .. math::
        b_j = \\begin{cases}
            \\frac{1}{(1 - \\sqrt{\\alpha_1})n} & \\text{if } j = 1 \\\\
            \\frac{2j - 4j\\alpha_2 + 2\\alpha_2 \\pm
             \\sqrt{2\\alpha_2}}{2n(1 - 2\\alpha_2)} & \\text{if } 1 < j < J\\\\
             1 & \\text{if } j = J
        \\end{cases}

    .. math::
        c_j = \\begin{cases}
            0 & \\text{if } j = 1 \\\\
            \\frac{a + b}{2} & \\text{if } 1 < j  < J\\\\
            1 & \\text{if } j = J
        \\end{cases}

    The value of :math:`\\alpha_1`, that represents the error for the first class,
    is computed as follows:

    .. math::
        \\alpha_1 = \\left(\\frac{1 - \\sqrt{1 - 4(1 - 2\\alpha_2)(2\\alpha_2 -
          \\sqrt{2\\alpha_2})}}{2}\\right)^2

    The value of :math:`\\alpha_3`, that represents the error for the last class,
    is computed as follows:

    .. math::
        \\alpha_3 = \\left(\\frac{1 -
        \\sqrt{1 - 4\\left(\\frac{J - 1}{J}\\right)^2(1 - 2\\alpha_2)(\\sqrt{2\\alpha_2}
          (-1 + \\sqrt{2\\alpha_2}))}}{2}\\right)^2


    The value of :math:`\\alpha_2` is given by the user.

    Parameters
    ----------
    J : int
        Number of classes or splits (:math:`J`).
    alpha2 : float, optional, default=0.01
        Value of the :math:`\\alpha_2` parameter.
    verbose : int, optional, default=0
        Verbosity level.

    Raises
    ------
    ValueError
        If ``J`` is not a positive integer greater than 1.
        If ``alpha2`` is not a float between 0 and 1.

    Returns
    -------
    probs : 2d array-like of shape (J, J)
        Matrix of probabilities where each row represents the true class
        and each column the probability for class j.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_triangular_soft_labels
    >>> get_triangular_soft_labels(5)
    array([[0.98845494, 0.01154505, 0.        , 0.        , 0.        ],
           [0.01      , 0.98      , 0.01      , 0.        , 0.        ],
           [0.        , 0.01      , 0.98      , 0.01      , 0.        ],
           [0.        , 0.        , 0.01      , 0.98      , 0.01      ],
           [0.        , 0.        , 0.        , 0.00505524, 0.99494475]])
    """

    if J < 2 or not isinstance(J, int):
        raise ValueError(f"{J=} must be a positive integer greater than 1")

    if alpha2 < 0 or alpha2 > 1:
        raise ValueError(f"{alpha2=} must be a float between 0 and 1")

    if verbose >= 1:
        print(f"Computing triangular probabilities for {J=} and {alpha2=}...")

    def compute_alpha1(alpha2):
        c_minus = (1 - 2 * alpha2) * (2 * alpha2 - math.sqrt(2 * alpha2))

        return pow((1 - math.sqrt(1 - 4 * c_minus)) / 2, 2)

    def compute_alpha3(alpha2):
        c1 = (
            pow((J - 1) / J, 2)
            * (1 - 2 * alpha2)
            * (math.sqrt(2 * alpha2) * (-1 + math.sqrt(2 * alpha2)))
        )

        return pow((1 - math.sqrt(1 - 4 * c1)) / 2, 2)

    alpha1 = compute_alpha1(alpha2)
    alpha3 = compute_alpha3(alpha2)

    if verbose >= 1:
        print(f"{alpha1=}, {alpha2=}, {alpha3=}")

    def b1(J):
        return 1.0 / ((1.0 - math.sqrt(alpha1)) * J)

    def aj(J, j):
        num1 = 2.0 * j - 2 - 4 * j * alpha2 + 2 * alpha2
        num2 = math.sqrt(2 * alpha2)
        den = 2.0 * J * (1 - 2 * alpha2)

        max_value = (j - 1.0) / J

        # +-
        return (
            (num1 + num2) / den
            if (num1 + num2) / den < max_value
            else (num1 - num2) / den
        )

    def bj(J, j):
        num1 = 2.0 * j - 4 * j * alpha2 + 2 * alpha2
        num2 = math.sqrt(2 * alpha2)
        den = 2.0 * J * (1 - 2 * alpha2)

        min_value = j / J

        # +-
        return (
            (num1 + num2) / den
            if (num1 + num2) / den > min_value
            else (num1 - num2) / den
        )

    def aJ(J):
        aJ_plus = 1.0 + 1.0 / (J * (math.sqrt(alpha3) - 1.0))
        aJ_minus = 1.0 + 1.0 / (-J * (math.sqrt(alpha3) - 1.0))
        return aJ_plus if aJ_plus > 0.0 else aJ_minus

    if verbose >= 3:
        print(f"{b1(J)=}, {aJ(J)=}, {aj(J, 1)=}, {bj(J,1)=}")
        for i in range(1, J + 1):
            print(f"{i=}  {aj(J, i)=}, {bj(J,i)=}")

    intervals = get_intervals(J)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    for j in range(1, J + 1):
        j_probs = []
        if j == 1:
            a = 0.0
            b = b1(J)
            c = 0.0
        elif j == J:
            a = aJ(J)
            b = 1.0
            c = 1.0
        else:
            a = aj(J, j)
            b = bj(J, j)
            c = (a + b) / 2.0

        if verbose >= 1:
            print(f"Class: {j}, {a=}, {b=}, {c=}, (j-1)/J={(j-1)/J}, (j/J)={j/J}")

        for interval in intervals:
            j_probs.append(
                triangular_cdf(interval[1], a, b, c)
                - triangular_cdf(interval[0], a, b, c)
            )
            if verbose >= 2:
                print(f"\tinterval: {interval}, prob={j_probs[-1]}")

        probs.append(j_probs)

    return np.array(probs)



from typing import Callable

import torch
from torch import nn


class ResNetOrdinalFullyConnected(nn.Module):
    """
    ResNetOrdinalFullyConnected implements the ordinal fully connected layer

    Parameters
    ----------
    input_size: int
        Input size
    num_classes: int
        Number of classes
    """

    classifiers: nn.ModuleList

    def __init__(self, input_size: int, num_classes: int):
        super(ResNetOrdinalFullyConnected, self).__init__()
        self.classifiers = nn.ModuleList(
            [nn.Linear(input_size, 1) for _ in range(num_classes - 1)]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        """
        xs = [classifier(x) for classifier in self.classifiers]
        x = torch.cat(xs, dim=1)
        x = torch.sigmoid(x)
        return x


class VGGOrdinalFullyConnected(nn.Module):
    """
    VGGOrdinalFullyConnected implements the ordinal fully connected layer

    Parameters
    ----------
    input_size: int
        Input size
    num_classes: int
        Number of classes
    activation_function: Callable[[], nn.Module]
        Activation function
    """

    classifiers: nn.ModuleList

    def __init__(
        self,
        input_size: int,
        num_classes: int,
        activation_function: Callable[[], nn.Module],
    ):
        super(VGGOrdinalFullyConnected, self).__init__()
        hidden_size = 4096 // (num_classes - 1)
        self.classifiers = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Linear(input_size, hidden_size),
                    activation_function(),
                    nn.Dropout(),
                    nn.Linear(hidden_size, hidden_size),
                    activation_function(),
                    nn.Dropout(),
                    nn.Linear(hidden_size, 1),
                )
                for _ in range(num_classes - 1)
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        """
        xs = [classifier(x) for classifier in self.classifiers]
        x = torch.cat(xs, dim=1)
        x = torch.sigmoid(x)
        return x



import numpy as np
from scipy.special import gamma, hyp2f1

from .utils import get_intervals


def beta_func(a, b):
    """Compute the beta function.

    Parameters
    ----------
    a : float
            First parameter.
    b : float
            Second parameter.

    Returns
    -------
    beta_func : float
            The value of the beta function.
    """

    if a <= 0:
        raise ValueError(f"{a=} can not be negative")

    if b <= 0:
        raise ValueError(f"{b=} can not be negative")

    return (gamma(a) * gamma(b)) / gamma(a + b)


def beta_dist(x, p, q, a=1.0):
    """Compute the beta distribution function for value x with parameters p, q and a.

    Parameters
    ----------
    x : float
            Value to compute the distribution function for.
    p : float
            First shape parameter (:math:`p > 0`).
    q : float
            Second shape parameter (:math:`q > 0`).
    a : float, default=1.0
            Scaling parameter (:math:`a > 0`).

    Returns
    -------
    beta: float
            The value of the beta distribution function.
    """

    if x < 0 or x > 1:
        raise ValueError(f"{x=} must be in the interval [0,1]")

    if a <= 0:
        raise ValueError(f"{a=} can not be negative or 0")

    return (x ** (a * p)) / (p * beta_func(p, q)) * hyp2f1(p, 1 - q, p + 1, x**a)


def _get_beta_soft_label(J, p, q, a=1.0):
    """Get soft labels from a beta distribution :math:`B(p,q,a)` for ``J`` splits.
    The :math:`[0,1]` interval is split into ``J`` intervals and the probability for
    each interval is computed as the difference between the value of the distribution
    function in the upper limit of the interval and the value of the distribution
    function in the lower limit of the interval. Thus, the probability for class ``j``
    is computed as :math:`B(p,q,a)(j/J) - B(p,q,a)((j-1)/J)`.

    Parameters
    ----------
    J : int
            Number of classes or splits.
    p : float
            First shape parameter (:math:`p > 0`).
    q : float
            Second shape parameter (:math:`q > 0`).
    a : float, default=1.0
            Scaling parameter (:math:`a > 0`).

    Raises
    ------
    ValueError
            If ``J`` is not a positive integer, if ``p`` is not positive, if ``q`` is
            not positive or if ``a`` is not positive.

    Returns
    -------
    probs: list
            List of ``J`` elements that represent the probability associated with each
            class or split.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_beta_soft_labels
    >>> get_beta_soft_labels(3, 2, 3)
    [0.4074074080000002, 0.48148148059259255, 0.11111111140740726]
    >>> get_beta_soft_labels(5, 5, 1, a=2)
    [1.0240000307200007e-07, 0.00010475520052121611, 0.005941759979320316,
      0.10132756401484902, 0.8926258084053067]
    """

    if J < 2 or not isinstance(J, int):
        raise ValueError(f"{J=} must be a positive integer greater than 1")

    if p <= 0:
        raise ValueError(f"{p=} must be positive")

    if q <= 0:
        raise ValueError(f"{q=} must be positive")

    if a <= 0:
        raise ValueError(f"{a=} must be positive")

    intervals = get_intervals(J)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    for interval in intervals:
        probs.append(beta_dist(interval[1], p, q, a) - beta_dist(interval[0], p, q, a))

    return probs


# Params [p,q,a] for beta distribution
_beta_params_sets = {
    "standard": {
        3: [[1, 4, 1], [4, 4, 1], [4, 1, 1]],
        4: [[1, 6, 1], [6, 10, 1], [10, 6, 1], [6, 1, 1]],
        5: [[1, 8, 1], [6, 14, 1], [12, 12, 1], [14, 6, 1], [8, 1, 1]],
        6: [[1, 10, 1], [7, 20, 1], [15, 20, 1], [20, 15, 1], [20, 7, 1], [10, 1, 1]],
        7: [
            [1, 12, 1],
            [7, 26, 1],
            [16, 28, 1],
            [24, 24, 1],
            [28, 16, 1],
            [26, 7, 1],
            [12, 1, 1],
        ],
        8: [
            [1, 14, 1],
            [7, 31, 1],
            [17, 37, 1],
            [27, 35, 1],
            [35, 27, 1],
            [37, 17, 1],
            [31, 7, 1],
            [14, 1, 1],
        ],
        9: [
            [1, 16, 1],
            [8, 40, 1],
            [18, 47, 1],
            [30, 47, 1],
            [40, 40, 1],
            [47, 30, 1],
            [47, 18, 1],
            [40, 8, 1],
            [16, 1, 1],
        ],
        10: [
            [1, 18, 1],
            [8, 45, 1],
            [19, 57, 1],
            [32, 59, 1],
            [45, 55, 1],
            [55, 45, 1],
            [59, 32, 1],
            [57, 19, 1],
            [45, 8, 1],
            [18, 1, 1],
        ],
        11: [
            [1, 21, 1],
            [8, 51, 1],
            [20, 68, 1],
            [34, 73, 1],
            [48, 69, 1],
            [60, 60, 1],
            [69, 48, 1],
            [73, 34, 1],
            [68, 20, 1],
            [51, 8, 1],
            [21, 1, 1],
        ],
        12: [
            [1, 23, 1],
            [8, 56, 1],
            [20, 76, 1],
            [35, 85, 1],
            [51, 85, 1],
            [65, 77, 1],
            [77, 65, 1],
            [85, 51, 1],
            [85, 35, 1],
            [76, 20, 1],
            [56, 8, 1],
            [23, 1, 1],
        ],
        13: [
            [1, 25, 1],
            [8, 61, 1],
            [20, 84, 1],
            [36, 98, 1],
            [53, 100, 1],
            [70, 95, 1],
            [84, 84, 1],
            [95, 70, 1],
            [100, 53, 1],
            [98, 36, 1],
            [84, 20, 1],
            [61, 8, 1],
            [25, 1, 1],
        ],
        14: [
            [1, 27, 1],
            [2, 17, 1],
            [5, 23, 1],
            [9, 27, 1],
            [13, 28, 1],
            [18, 28, 1],
            [23, 27, 1],
            [27, 23, 1],
            [28, 18, 1],
            [28, 13, 1],
            [27, 9, 1],
            [23, 5, 1],
            [17, 2, 1],
            [27, 1, 1],
        ],
    }
}


def get_beta_soft_labels(J, params_set="standard"):
    """Get soft labels for each of the ``J`` classes using a beta distributions and
    the parameter defined in the ``params_set`` as described in :footcite:t:`vargas2022unimodal`.

    Parameters
    ----------
    J : int
            Number of classes or splits.

    params_set : str, default='standard'
            The set of parameters of the beta distributions employed to generate the
            soft labels. It has to be one of the keys in the ``_beta_params_sets``
            dictionary.

    Raises
    ------
    ValueError
            If ``J`` is not a positive integer or if ``params_set`` is not a valid key.

    Returns
    -------
    probs: list
            List of ``J`` elements where each elements is also a list of ``J`` elements.
            Each inner list represents the soft label of class ``j`` for each of the
            ``J`` classes. For example: ``probs[j]`` is the soft label for class ``j``.
            Then, ``probs[j][k]`` is the probability of assigning class ``k`` to the
            instance that belongs to class ``j``.

    Example
    -------
    >>> from dlordinal.soft_labelling import get_beta_softlabels
    >>> get_beta_softlabels(3)
    [[0.802469132197531, 0.18518518474074064, 0.01234567906172801],
    [0.17329675405578432, 0.6534064918884313, 0.1732967540557846],
    [0.012345679061728405, 0.1851851847407408, 0.8024691321975309]]
    >>> get_beta_softlabels(5)
    [[0.8322278330066323, 0.15097599903815717, 0.01614079995258888,
    0.0006528000025611824, 2.5600000609360407e-06], [0.16306230596685573,
    0.6740152119811103, 0.15985465217901373, 0.003067150177798128,
    6.796952229937148e-07], [0.0005973937258183486, 0.16304604740785777,
    0.6727131177326486, 0.16304604740785367, 0.000597393725820794],
    [6.796952207410873e-07, 0.0030671501777993896, 0.15985465217901168,
    0.6740152119811109, 0.16306230596685678], [2.560000061440001e-06,
    0.0006528000025600005, 0.01614079995258879, 0.1509759990381568,
    0.8322278330066332]]
    """

    if J <= 0:
        raise ValueError(f"{J=} must be a positive integer")

    if params_set not in _beta_params_sets:
        raise ValueError(f"Invalid params_set: {params_set}")

    params = _beta_params_sets[params_set]
    return np.array([_get_beta_soft_label(J, p, q, a) for (p, q, a) in params[J]])
